#include <stdio.h>
#include <motelib/system.h>
#include <motelib/led.h>
#include <motelib/timer.h>
#include <motelib/radio.h>
#include <motelib/uart.h>
#include <motelib/sensor.h>
#include <motelib/actor.h>

#define HEAT          0
#define LIGHT         1
#define BASE_STATION  0
#define SOURCE_ADDR   24
#define TABLE_SIZE    5
#define NO_OF_TYPE    2
#define INTEREST_TYPE 150
#define DATA_TYPE     151
#define ACK_TYPE      152

typedef struct interest_t {
  uint8_t seq_no;
  uint8_t data_type;
} interest;

typedef struct ack_t {
  uint8_t ack_no;  
  uint8_t data_type;
} ack;

typedef struct data_t {
  uint8_t seq_no;
  uint8_t data_type;
  uint16_t info;
} data;

typedef struct row_t {
  uint16_t source;
  uint8_t gradient;
} row;


Timer t1, moveTime, steerTime;
uint8_t addr_lo, addr_hi;
uint8_t seq_no[2];
uint8_t data_seq[2];
uint8_t ack_no[2];

//Neighbour table
row table[NO_OF_TYPE][TABLE_SIZE];


void requestHeat(Timer *t);
void fadeGradient(Timer *t);
void receive(Address source,MessageType type,void *message,uint8_t len);
void lightReady(uint16_t value);
void tempReady(uint16_t value);
void senseLight();
void senseTemp();


void requestHeat(Timer *t)
{
  interest flood;
  flood.seq_no = seq_no[HEAT];
  flood.data_type = HEAT;
  	 
  seq_no[HEAT]++;
  radioRequestTx(BROADCAST_ADDR, INTEREST_TYPE, &flood, sizeof(flood), NULL);
}

void requestLight(Timer *t)
{
  interest flood;
  flood.seq_no = seq_no[LIGHT];
  flood.data_type = LIGHT;
     
  seq_no[LIGHT]++;
  radioRequestTx(BROADCAST_ADDR, INTEREST_TYPE, &flood, sizeof(flood), NULL);
}

void fadeGradient(Timer *t) {
  int i, j;
  for (i = 0; i<NO_OF_TYPE; i++) {
   for (j = 0; j<TABLE_SIZE; j++) {
     if (table[i][j].gradient > 8)
        table[i][j].gradient-=8;
      else
        table[i][j].gradient = 0;
    }
  }
}
void onStopMoving() {
  actorSetState(ACTOR_5, 0);
  actorSetState(ACTOR_3, 0);
}
void OnReceiveCommand(char command, uint8_t interval) {
  if (command == 'F') {
    actorSetState(ACTOR_5, 0);
    actorSetState(ACTOR_3, 1);
    timerStart(&moveTime, TIMER_ONESHOT, interval*1000, onStopMoving);
  } else if (command == 'B') {
     actorSetState(ACTOR_3, 0);
    actorSetState(ACTOR_5, 1);
    timerStart(&moveTime, TIMER_ONESHOT, interval*1000, onStopMoving);
  }
}
void showHeat(uint16_t x)
{
  //<260 blue
  //260-512 light blue
  //513-640 green
  //641-790 yellow
  //791 up red

  //red PC1
  //blue PC3
  //green PC5
  if(x > 580)
  {
    actorSetState(ACTOR_1,1);
    actorSetState(ACTOR_3,0);
    actorSetState(ACTOR_5,0);
  }
  else if(x >550)
  {
    actorSetState(ACTOR_1,1);
    actorSetState(ACTOR_3,0);
    actorSetState(ACTOR_5,1);
  }
  else if(x >520)
  {
    actorSetState(ACTOR_1,0);
    actorSetState(ACTOR_3,0);
    actorSetState(ACTOR_5,1);
  }
  else if(x >500)
  {
    actorSetState(ACTOR_1,0);
    actorSetState(ACTOR_3,1);
    actorSetState(ACTOR_5,1);
  }
  else
  {
    actorSetState(ACTOR_1,0);
    actorSetState(ACTOR_3,1);
    actorSetState(ACTOR_5,0);
  }

}
void receive(Address source,MessageType type,void *message,uint8_t len)
{
  if (type == INTEREST_TYPE) {
      int i = 0;
      interest msg = *(interest*)message;
      if ((int8_t)(msg.seq_no - seq_no[msg.data_type]) <= 0) return;
      seq_no[msg.data_type] = msg.seq_no;
      
      if(getAddress() == SOURCE_ADDR && msg.data_type == HEAT){
        senseTemp();
      }
      else if(getAddress() != BASE_STATION){
        interest flood = msg;
        radioRequestTx(BROADCAST_ADDR, INTEREST_TYPE, &flood, sizeof(flood), NULL);
      }
      
      //New record
      for (i = 0; i<TABLE_SIZE; i++) {
        if (table[msg.data_type][i].source == source)
          return;
      }
      for (i = 0; i<TABLE_SIZE; i++) {
        if (table[msg.data_type][i].source == 65535) {
          row newParent;
          newParent.source = source;
          newParent.gradient = 0;
        
          table[msg.data_type][i] = newParent;
          return;
        }
      }
    
  }
  else if (type == DATA_TYPE) {
    if (getAddress() == BASE_STATION) {
      data msg = *(data*)message;
      if ((int8_t)(msg.seq_no - data_seq[HEAT]) <= 0) return;
      //printf("src = %d msg seq = %d data_seq=%d\n",source,msg.seq_no,data_seq[msg.data_type]);
      if(msg.data_type == HEAT)
      {
        data_seq[HEAT] = msg.seq_no;
        printf("seq %d data = %d\n",data_seq[HEAT],msg.info);
        //showHeat(msg.info);
        // ack?
        ack newAck;
        ack_no[HEAT] = msg.seq_no+1;
        newAck.ack_no = ack_no[HEAT];
        newAck.data_type = msg.data_type;
        radioRequestTx(source, ACK_TYPE, &newAck, sizeof(newAck), NULL);
        return ;
      }
    } 
    if (getAddress() == SOURCE_ADDR) {
      data msg = *(data*)message;
      //printf("src = %d msg seq = %d data_seq=%d\n",source,msg.seq_no,data_seq);
      if(msg.data_type == LIGHT)
      {
        if ((int8_t)(msg.seq_no - data_seq[LIGHT]) <= 0) return;
        data_seq[LIGHT] = msg.seq_no;
        //printf("seq %d data = %d\n",data_seq[LIGHT],msg.info);
        //OnReceiveCommand(msg.info>>8,msg.info&255);
        printf("command = %c time = %d\n",msg.info>>8,msg.info&255);
        //showHeat(msg.info);
        // ack?
        ack newAck;
        ack_no[LIGHT] = msg.seq_no+1;
        newAck.ack_no = ack_no[LIGHT];
        newAck.data_type = msg.data_type;
        radioRequestTx(source, ACK_TYPE, &newAck, sizeof(newAck), NULL);
        return ;
      }
    } else {
      int i = 0;
      uint8_t max = 0;
      data msg = *(data*)message;
      if ((int8_t)(msg.seq_no - data_seq[msg.data_type]) <= 0) return;
      data_seq[msg.data_type] = msg.seq_no;
      data toSend = msg;
      
      //Find max gradient
      for (i = 0; i<TABLE_SIZE; i++) {
       if (table[msg.data_type][i].gradient > max)
         max = table[msg.data_type][i].gradient;

      }
      //Send to max gradient
      for (i = 0; i<TABLE_SIZE; i++) {
        
       if (table[msg.data_type][i].gradient == max  && table[msg.data_type][i].source!=65535){
        
        radioRequestTx(table[msg.data_type][i].source, DATA_TYPE, &toSend, sizeof(toSend), NULL);
        //printf("%d, %d, %d\n",i, table[msg.data_type][i].source, table[msg.data_type][i].gradient);
        }
      }
      
      //Acknowledge
      ack newAck;
      ack_no[msg.data_type] = msg.seq_no+1;
      newAck.ack_no = ack_no[msg.data_type];
      newAck.data_type = msg.data_type;
      radioRequestTx(source, ACK_TYPE, &newAck, sizeof(newAck), NULL);
    }
  }
  else if (type == ACK_TYPE) {
    int i = 0;
    ack msg = *(ack*)message;
    //printf("ACK SCR = %d msg.ack_no %d ack_no %d type: %d\n", source,msg.ack_no,ack_no[msg.data_type],msg.data_type);
    if ((int8_t)(msg.ack_no - ack_no[msg.data_type]) < 0) return;
    ack_no[msg.data_type] = msg.ack_no+1;
    for (i = 0; i<TABLE_SIZE; i++) {
       if (table[msg.data_type][i].source == source) {
         if (table[msg.data_type][i].gradient < 239)
          table[msg.data_type][i].gradient+=16;
       }
     }
  }
}

void lightReady(uint16_t value)
{
  int i = 0;
  uint8_t max = 0;
  data toSend;
  toSend.seq_no = data_seq[LIGHT];
  toSend.info = value;
  toSend.data_type = LIGHT;
  //find max gradient
  for (i = 0; i<TABLE_SIZE; i++) {
    if (table[LIGHT][i].gradient > max)
      max =  table[LIGHT][i].gradient;
  }
  //send to max gradient
  for (i = 0; i<TABLE_SIZE; i++) {
    if (table[LIGHT][i].gradient == max && table[LIGHT][i].source!=65535)
      radioRequestTx(table[LIGHT][i].source, DATA_TYPE, &toSend, sizeof(toSend), NULL);
  }
  data_seq[LIGHT]++;
}

void tempReady(uint16_t value) {
  int i = 0;
  uint8_t max = 0;
  data toSend;
  toSend.seq_no = data_seq[HEAT];
  toSend.info = value;
  toSend.data_type = HEAT;
  //find max gradient
  
  for (i = 0; i<TABLE_SIZE; i++) {
    if (table[HEAT][i].gradient > max)
      max =  table[HEAT][i].gradient;
  }
  //send to max gradient
  for (i = 0; i<TABLE_SIZE; i++) {
    if (table[HEAT][i].gradient == max && table[HEAT][i].source!=65535){
      radioRequestTx(table[HEAT][i].source, DATA_TYPE, &toSend, sizeof(toSend), NULL);
      
    }
  }
  data_seq[HEAT]++;
}

void senseLight()
{
  //actorSetState(ACTOR_0, 1);
  //sensorRequestAnalog(SENSOR_1, lightReady);
  //lightReady(2);
}

void senseTemp()
{
  actorSetState(ACTOR_0, 1);
  //sensorRequestAnalog(SENSOR_2, tempReady);
  tempReady(3);
}

void ready(uint8_t markerFound)
{
  uint16_t i,l=uartInputLen();
  uint8_t command,sec;
  char c,s;
  for(i=0;i<l;i++)
  {
    c = uartReadByte();
    if(c>='A' && c<='Z'){
      printf("%c",c);
      switch(c){
        case 'F': command = 'F';break;
        case 'B': command = 'B';break;
        case 'L': command = 'L';break;
        case 'R': command = 'R';break;
        default : uartFlushInput();uartWatchInputMarker('\r',256,ready); return; break;
      }
      s = uartReadByte();
      printf("%c",s);
      if(s >= '0' && s <='9')
      {
        sec=s-'0';
      }
      else{
        uartFlushInput();uartWatchInputMarker('\r',256,ready); return;
      }
      //printf("?????\n");
      uint16_t out = command;
      out = (out<<8)+sec;
      lightReady(out);
    }
  }
  uartWatchInputMarker('\r',256,ready);
}

Timer t2,t3;

void boot()
{
  addr_lo = getAddress()%256;
  addr_hi = getAddress()/256;
  
  ledSetValue(1);
  timerCreate(&t1);
  timerCreate(&t2);
  timerCreate(&t3);
  timerCreate(&moveTime);
  timerCreate(&steerTime);
  data_seq[LIGHT]=0;
  data_seq[HEAT]=0;
  //Init table
  int i,j;
  for (i = 0; i<NO_OF_TYPE; i++) {
   for (j = 0; j<TABLE_SIZE; j++) {
      row empty;
      empty.source = -1;
      empty.gradient = 0;
      table[i][j] = empty;
    }
  }
  if (getAddress() == BASE_STATION) {
    seq_no[HEAT]++;
    //timerStart(&t3, TIMER_PERIODIC, 11000, senseLight);
    timerStart(&t2, TIMER_PERIODIC, 3900, requestHeat);
    timerStart(&t1, TIMER_PERIODIC, 10000, fadeGradient);
    ledSetValue(1);
    uartWatchInputMarker('\r',256,ready);
  } else if (getAddress() == SOURCE_ADDR) {
    seq_no[LIGHT]++;
    data_seq[LIGHT]++;
    timerStart(&t2, TIMER_PERIODIC, 7000, requestLight);
  	timerStart(&t1, TIMER_PERIODIC, 10000, fadeGradient);
    timerStart(&t3, TIMER_ONESHOT, 500, requestLight);
    
    ledSetValue(2);
  } else {
  	timerStart(&t1, TIMER_PERIODIC, 10000, fadeGradient);
    ledSetValue(4);
  }
  
  uartEnable(true, true);
 
   
  radioSetRxHandler(receive);
}

